from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.db import transaction

from codelist.models import KonovaCode
from codelist.settings import CODELIST_BIOTOPES_ID, CODELIST_BIOTOPES_EXTRA_CODES_ID
from compensation.models import Compensation, CompensationState
from intervention.models import Responsibility, Handler, Intervention
from konova.management.commands.kspMigrater.base_migrater import BaseMigrater
from konova.models import Deadline, DeadlineType


class CompensationMigrater(BaseMigrater):

    def migrate(self):
        cursor = self.db_connection.cursor()
        empty_str = "''"
        cursor.execute(
            'select '
                'om."KENNUNG", '
                'linf."OBJBEZ", '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geomf.the_geom,4326)), 3))) as geomf, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geoml.the_geom,4326)), 2))) as geoml, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(ST_Transform(geomp.the_geom,4326)), 1))) as geomp, '
                'linf."Bemerkung" '
            'from "OBJ_MASTER" om '
                'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
                'left join kom on om."GISPADID"=kom.gispadid '
                'left join geometry_f geomf on om."GISPADID"=geomf.gispadid '
                'left join geometry_l geoml on om."GISPADID"=geoml.gispadid '
                'left join geometry_p geomp on om."GISPADID"=geomp.gispadid '
                'left join "Aufwertung" auf on om."GISPADID"=auf."GISPADID" '
            'where '
                'om."OKL"=7730080 and '
                'om.archiv=false and '
                f'(auf."Infos" is null or auf."Infos"={empty_str}) and '
                'om.nicht_vollstaendig=0;'
        )

        all_koms = cursor.fetchall()
        len_all_koms = len(all_koms)
        num_processed = 0
        print(f"Migrate KOMs to compensations...")
        print(f"--Found {len_all_koms} entries. Process now...")
        unsuccessfull_compensations = {}
        for kom in all_koms:
            if num_processed % 500 == 0:
                print(f"----{num_processed}/{len_all_koms} processed")

            with transaction.atomic():
                kom_identifier = kom[0]
                kom_title = kom[1]
                kom_comment = kom[5]
                compensation = Compensation.objects.get_or_create(
                    identifier=kom_identifier
                )[0]

                compensation.title = kom_title
                compensation.comment = kom_comment
                compensation = self._migrate_geometry(compensation, kom)
                compensation = self._migrate_responsibility(compensation, kom)
                compensation = self._migrate_compensation_type(compensation, kom)
                compensation = self._migrate_states(compensation, kom)
                compensation = self._migrate_deadlines(compensation, kom)
                try:
                    compensation = self._migrate_interventions_reference(compensation, kom)
                    compensation.save()
                except ObjectDoesNotExist:
                    compensation.delete()
                    unsuccessfull_compensations[kom_identifier] = "EIV does not exist"
                num_processed += 1
        print("The following KOMs could not be migrated: ")
        for kom, val in unsuccessfull_compensations:
            print(kom)
        cursor.close()

    def _migrate_interventions_reference(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'ref."REFERENZ" '
            'from "OBJ_MASTER" om '
            'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
            'left join "REFERENZ" ref on om."GISPADID"=ref."GISPADID" '
            'where '
                'om."OKL"=7730080 and '
                f'om."KENNUNG"={kom_identifier};'
        )
        eivs = tmp_cursor.fetchall()
        len_eivs = len(eivs)
        if len_eivs != 1:
            raise AssertionError(f"{kom_identifier} has {len_eivs} EIVs!")
        eiv = eivs[0]
        try:
            intervention = Intervention.objects.get(
                identifier=eiv[0]
            )
        except ObjectDoesNotExist:
            raise ObjectDoesNotExist(f"{kom_identifier} could not find migrated {eiv}")
        compensation.intervention = intervention
        tmp_cursor.close()
        return compensation

    def _migrate_responsibility(self, compensation, kom):
        compensation.responsible = compensation.responsible or Responsibility.objects.create()
        return compensation

    def _migrate_states(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        before_states_sql = 'select ' \
                'b."Biotoptyp", ' \
                'b."Flaeche", ' \
                'b."PKEY" ' \
            'from "BtypHtyp" b ' \
                'left join "OBJ_MASTER" om on om."GISPADID"=b."GISPADID" ' \
            'where ' \
                f'om."KENNUNG"={kom_identifier}' 
        
        after_states_sql = 'select ' \
                'z."Zielbiotoptyp", ' \
                'z."Flaeche", ' \
                'z."PKEY" ' \
            'from "Zielbiotoptypen" z ' \
                'left join "OBJ_MASTER" om on om."GISPADID"=z."GISPADID" ' \
            'where ' \
                f'om."KENNUNG"={kom_identifier}' 
        z_code_sql = 'select z."Zusatzcode" from "Zusatzcodes" z where z."FKEY"={}'
        
        compensation = self._process_state_migration(compensation, before_states_sql, z_code_sql, compensation.before_states)
        compensation = self._process_state_migration(compensation, after_states_sql, z_code_sql, compensation.after_states)
        return compensation

    def _process_state_migration(self, compensation, sql, z_code_sql, state_manager):
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(sql)
        db_result = tmp_cursor.fetchall()

        for result in db_result:
            state_type = result[0]
            state_surface = result[1] or 0.0
            pkey_entry = f"'{result[2]}'"
            state_code = KonovaCode.objects.get(
                atom_id=state_type,
                code_lists__in=[CODELIST_BIOTOPES_ID]
            )
            tmp_cursor_z_code = self.db_connection.cursor()
            tmp_cursor_z_code.execute(z_code_sql.format(pkey_entry))
            z_code_results = tmp_cursor_z_code.fetchall()
            z_codes = []
            
            for z_code_result in z_code_results:
                z_code = KonovaCode.objects.filter(
                    atom_id=z_code_result[0],
                    code_lists__in=[CODELIST_BIOTOPES_EXTRA_CODES_ID],
                )
                z_codes += z_code
            tmp_cursor_z_code.close()

            try:
                _state_obj = state_manager.get(
                    biotope_type=state_code,
                    surface=state_surface
                )
            except ObjectDoesNotExist:
                _state_obj = CompensationState.objects.create(
                    biotope_type=state_code,
                    surface=state_surface,
                )
                state_manager.add(_state_obj)

            _state_obj.biotope_type_details.set(z_codes)

        tmp_cursor.close()
        return compensation

    def _migrate_compensation_type(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'mt.typ '
            'from "OBJ_MASTER" om '
                'left join massnahmetyp mt on om."GISPADID"=mt.gispadid '
            'where '
                f'om."KENNUNG"={kom_identifier}'
        )
        db_result = tmp_cursor.fetchall()
        if len(db_result) != 1:
            raise AssertionError(f"{kom_identifier} has no specification on compensation type (CEF, ...)")
        comp_type = db_result[0][0]
        if comp_type == 705816:
            # regular compensation, do nothing
            pass
        elif comp_type == 705815:
            compensation.is_cef = True
        elif comp_type == 705817:
            compensation.is_coherence_keeping = True
        elif comp_type == 154156555:
            compensation.is_coherence_keeping = True
            compensation.is_cef = True
        tmp_cursor.close()
        return compensation

    def _migrate_deadlines(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'f.umsetzungbis, '
                'f.unterhaltungbis, '
                'f.artunterhaltungbis, '
                'f.artumsetzungbis '
            'from kom k '
                'left join fristen f on k.fristen=f.id '
                'left join "OBJ_MASTER" om on k.gispadid=om."GISPADID" '
            'where '
                f'om."KENNUNG"={kom_identifier}'
        )
        db_results = tmp_cursor.fetchall()

        for result in db_results:
            finish_until = result[0]
            maintain_until = result[1]
            maintain_until_type = result[2]
            finish_until_type = result[3]

            if finish_until_type == 941256187:
                finish_until_type = "Termin endgültig"
            elif finish_until_type == 889700393:
                finish_until_type = "Termin vorläufig"
            else:
                finish_until_type = "Unbekannt ob Termin vorläufig oder endgültig"

            if maintain_until_type == 807238388:
                maintain_until_type = "Terminart 'bis zum'"
            elif maintain_until_type == 23925195:
                maintain_until_type = "Terminart 'dauerhaft'"
            elif maintain_until_type == 322729640:
                maintain_until_type = "Terminart 'nicht erforderlich'"
            else:
                maintain_until_type = "Unbekannte Terminart"

            if finish_until is not None:
                try:
                    deadline = compensation.deadlines.get(
                        type=DeadlineType.FINISHED,
                        date=finish_until,
                        comment=finish_until_type
                    )
                except ObjectDoesNotExist:
                    deadline = Deadline.objects.create(
                        type=DeadlineType.FINISHED,
                        date=finish_until,
                        comment=finish_until_type
                    )
                    compensation.deadlines.add(deadline)

            if maintain_until is not None:
                try:
                    deadline = compensation.deadlines.get(
                        type=DeadlineType.MAINTAIN,
                        date=maintain_until,
                        comment=maintain_until_type
                    )
                except ObjectDoesNotExist:
                    deadline = Deadline.objects.create(
                        type=DeadlineType.MAINTAIN,
                        date=maintain_until,
                        comment=maintain_until_type
                    )
                    compensation.deadlines.add(deadline)

        tmp_cursor.close()
        return compensation